@@ -1,12 +1,13 @@
 # Copyright (C) 2007 by Dr. Dieter Maurer, Eichendorffstr. 23, D-66386 St. Ingbert, Germany
 # see "LICENSE.txt" for details
-#       $Id: connection.py,v 1.2 2008/10/03 07:49:04 dieter Exp $
+#       $Id: connection.py,v 1.3 2013/07/03 11:27:49 dieter Exp $
 '''A historical connection.
 
 It delivers state current at a given time or serial.
 '''
 from time import gmtime
 from types import MethodType
+from inspect import getargspec
 
 from DateTime import DateTime
 from ZODB.Connection import Connection
@@ -14,7 +15,7 @@ from ZODB.TimeStamp import TimeStamp
 from ZODB.POSException import POSKeyError
 
 class Connection(Connection):
-  def __init__(self, baseConnection, timeOrSerial):
+  def __init__(self, baseConnection, timeOrSerial, before=False):
     '''creates a (historical) connection based on *baseConnection* that delivers state at or before *timeOrSerial*.
 
     *baseConnection* must be an (open) ZODB connection. Its 'db' is used
@@ -23,25 +24,36 @@ class Connection(Connection):
     *timeOrSerial* is either a time ('DateTime' or float/int) or a serial/transactionid ('string').
     '''
     ts = None
-    db = baseConnection.db()
-    try: super(Connection,self).__init__(); self._setDB(db)
-    except TypeError:
-      # maybe ZODB 3.6
-      super(Connection,self).__init__(db); self._opened=True
     if isinstance(timeOrSerial, str): ts = TimeStamp(timeOrSerial)
     if isinstance(timeOrSerial, DateTime): timeOrSerial = timeOrSerial.timeTime()
     if isinstance(timeOrSerial, (int, float)):
       ts = TimeStamp(*(gmtime(timeOrSerial)[:5] + (timeOrSerial%60,)))
-    self._storage = _HistoricalStorageWrapper(`ts`, self._storage)
+    db = baseConnection.db()
+    c = super(Connection, self).__init__
+    if "before" in getargspec(c).args:
+      # this is a modern ZODB which directly supports historical connections
+      if not before:
+        # we must slightly increment "ts" as we need the state
+        #   `at or before` while the connection gives us `before`.
+        ts = ts.laterThan(ts)
+      c(db, before=`ts`)
+      self.open()
+    else:
+      try: super(Connection,self).__init__(); self._setDB(db)
+      except TypeError:
+        # maybe ZODB 3.6
+        super(Connection,self).__init__(db); self._opened=True
+      self._storage = _HistoricalStorageWrapper(`ts`, self._storage)
 
 
 class _HistoricalStorageWrapper(object):
   '''ensure that we deliver state at or before a given serial.'''
   __base = None
 
-  def __init__(self, serial, baseStorage):
+  def __init__(self, serial, baseStorage, before):
     self.__base = baseStorage
     self.__serial = serial
+    self.__before = before
 
   _is_read_only = True
   def isReadOnly(self): return True
@@ -49,9 +61,11 @@ class _HistoricalStorageWrapper(object):
   def load(self, oid, version):
     serial = self.__serial
     S = self.__base
-    filter= lambda t, serial= serial: t['tid'] <= serial # this might need to become a picklable object, once ZEO supports 'filter'.
+    filter= before \
+            and (lambda t, serial= serial: t['tid'] < serial)\
+            or (lambda t, serial= serial: t['tid'] <= serial)
     if _hasSmartHistory(S):
-      ts= S.history(oid, None, 1, filter= filter,)
+      ts= hwrap(S.history)(oid, None, 1, filter= filter,)
     else:
       ts = HistoryFetcher(S, oid, None, last=1, filter=filter).next()
     if not ts: raise POSKeyError((oid, serial))
@@ -79,7 +93,7 @@ class HistoryFetcher:
     through *last* are returned.
     '''
     S = storage
-    history = S.history
+    history = hwrap(S.history)
     if _hasSmartHistory(S): self._smartHistory = history
     else: self._getRawHistory = history
     self._oid = oid
@@ -188,9 +202,22 @@ class _Filter:
     return fd
 
 
-###########################################################################
-## non standard ZODB extension handling
-try: from ZODB.BaseStorage import _hasSmartHistory
-except ImportError:
-  def _hasSmartHistory(storage): return 0
+def _hasSmartHistory(storage):
+  h = storage.history
+  return "filter" in getargspec(h).args
 
+## history wrapping
+#    Modern ZODB versions have discarded the `version` parameter to `history`
+#    For backward compatibility we still use it with this parameter
+#    Wrap as appropriate
+
+# Note: ZODB 3.10.5 is inconsistent with respect to this parameter
+#  "BaseStorage.history" still has it. We look at "FileStorage"
+#  to determine whether the parameter should be passed or not
+from ZODB.FileStorage import FileStorage
+
+if "version" in getargspec(FileStorage.history).args:
+  def hwrap(history): return history
+else:
+  def hwrap(history):
+    return lambda oid, version, *args, **kw: history(oid, *args, **kw)
